@@ -63,6 +63,9 @@ char gHostIP[32];
 unsigned int gPort;
 char gRUser[32];
 char gRPass[128];
+char gPreCommand[128];
+char gPerFileCommand[128];
+char gPostCommand[128];
 int gXtype;
 int gRecursive;
 int gDelete;
@@ -121,7 +124,7 @@ DIR *opendir(const char *const path)
 	char *dirpath;
 	size_t len;
 
-	p = (char *) malloc(sizeof(DIR));
+	p = (DIR *) malloc(sizeof(DIR));
 	if (p == NULL)
 		return NULL;
 	memset(p, 0, sizeof(DIR));
@@ -249,7 +252,13 @@ static void PerrorBox(const char *const whatFailed)
 
 /*VARARGS*/
 static void
-Log(int uiShow, const char *const fmt, ...)
+Log(
+#if defined(WIN32) || defined(_WINDOWS)
+		int uiShow,
+#else
+		int UNUSED(uiShow),
+#endif
+		const char *const fmt, ...)
 {
 	va_list ap;
 	struct tm *ltp;
@@ -295,6 +304,8 @@ Log(int uiShow, const char *const fmt, ...)
 		}
 		YieldUI(2);
 	}
+#else
+	LIBNCFTP_USE_VAR(uiShow);
 #endif
 }	/* Log */
 
@@ -397,14 +408,14 @@ OpenLog(void)
 {
 	FILE *fp;
 	char pathName[256];
-	struct stat st;
+	struct Stat st;
 	const char *openMode;
 
 	CloseLog();
 	(void) OurDirectoryPath(pathName, sizeof(pathName), kSpoolLog);
 
 	openMode = gLogOpenMode;
-	if ((stat(pathName, &st) == 0) && (st.st_size > 200000L)) {
+	if ((Stat(pathName, &st) == 0) && (st.st_size > 200000L)) {
 		/* Prevent the log file from growing forever. */
 		openMode = FOPEN_WRITE_TEXT;
 	}
@@ -505,6 +516,9 @@ PreInit(void)
 	ZeroMemory(gStatusText, sizeof(gStatusText));
 #else
 	gIsTTY = ((isatty(2) != 0) && (getppid() > 1)) ? 1 : 0;
+#endif
+#ifdef SIGPOLL
+	NcSignal(SIGPOLL, (FTPSigProc) SIG_IGN);
 #endif
 	InitUserInfo();
 
@@ -515,6 +529,25 @@ PreInit(void)
 	(void) OurDirectoryPath(gSpoolDir, sizeof(gSpoolDir), kSpoolDir);
 	(void) signal(SIGINT, SigExit);
 	(void) signal(SIGTERM, SigExit);
+#if defined(WIN32) || defined(_WINDOWS)
+#else
+	(void) signal(SIGSEGV, SigExit);
+	(void) signal(SIGBUS, SigExit);
+	(void) signal(SIGFPE, SigExit);
+	(void) signal(SIGILL, SigExit);
+#if SIGIOT != SIGABRT
+	(void) signal(SIGIOT, SigExit);
+#endif
+#ifdef SIGEMT
+	(void) signal(SIGEMT, SigExit);
+#endif
+#ifdef SIGSYS
+	(void) signal(SIGSYS, SigExit);
+#endif
+#ifdef SIGSTKFLT
+	(void) signal(SIGSTKFLT, SigExit);
+#endif
+#endif
 }	/* PreInit */
 
 
@@ -558,6 +591,9 @@ PrePrintItem(void)
 	gLDir[0] = '\0';
 	gRFile[0] = '\0';
 	gLFile[0] = '\0';
+	gPreCommand[0] = '\0';
+	gPerFileCommand[0] = '\0';
+	gPostCommand[0] = '\0';
 
 	line[sizeof(line) - 1] = '\0';
 	while (fgets(line, sizeof(line) - 1, fp) != NULL) {
@@ -602,6 +638,12 @@ PrePrintItem(void)
 			(void) STRNCPY(gRFile, tok2);
 		} else if (strcmp(tok1, "local-file") == 0) {
 			(void) STRNCPY(gLFile, tok2);
+		} else if (strcmp(tok1, "pre-command") == 0) {
+			(void) STRNCPY(gPreCommand, tok2);
+		} else if (strcmp(tok1, "per-file-command") == 0) {
+			(void) STRNCPY(gPerFileCommand, tok2);
+		} else if (strcmp(tok1, "post-command") == 0) {
+			(void) STRNCPY(gPostCommand, tok2);
 		}	/* else, unknown command. */
 	}
 	(void) fclose(fp);
@@ -671,6 +713,7 @@ DoItem(int iType)
 	int needOpen;
 	int result;
 	int n;
+	int cdflags;
 	
 	fp = fopen(gMyItemPath, FOPEN_READ_TEXT);
 	if (fp == NULL) {
@@ -692,6 +735,9 @@ DoItem(int iType)
 	gLDir[0] = '\0';
 	gRFile[0] = '\0';
 	gLFile[0] = '\0';
+	gPreCommand[0] = '\0';
+	gPerFileCommand[0] = '\0';
+	gPostCommand[0] = '\0';
 
 	line[sizeof(line) - 1] = '\0';
 	while (fgets(line, sizeof(line) - 1, fp) != NULL) {
@@ -737,6 +783,12 @@ DoItem(int iType)
 			(void) STRNCPY(gRFile, tok2);
 		} else if (strcmp(tok1, "local-file") == 0) {
 			(void) STRNCPY(gLFile, tok2);
+		} else if (strcmp(tok1, "pre-command") == 0) {
+			(void) STRNCPY(gPreCommand, tok2);
+		} else if (strcmp(tok1, "per-file-command") == 0) {
+			(void) STRNCPY(gPerFileCommand, tok2);
+		} else if (strcmp(tok1, "post-command") == 0) {
+			(void) STRNCPY(gPostCommand, tok2);
 		}	/* else, unknown command. */
 	}
 	(void) fclose(fp);
@@ -757,6 +809,7 @@ DoItem(int iType)
 		}
 	}
 
+	cdflags = kChdirOneSubdirAtATime;
 	if (gOperation == 'G') {
 		if (gRecursive != 0) {
 			if (gRFile[0] == '\0') {
@@ -778,6 +831,7 @@ DoItem(int iType)
 			}
 		}
 	} else if (gOperation == 'P') {
+		cdflags = kChdirOneSubdirAtATime|kChdirAndMkdir;
 		if (gRecursive != 0) {
 			if (gLFile[0] == '\0') {
 				Log(0, "batch file parameter missing: %s.\n", "local-file");
@@ -844,6 +898,7 @@ DoItem(int iType)
 	}
 
 	if (needOpen != 0) {
+		(void) AdditionalCmd(&gConn, gPostCommand, NULL);
 		(void) FTPCloseHost(&gConn);
 		if (FTPInitConnectionInfo(&gLib, &gConn, kDefaultFTPBufSize) < 0) {
 			/* Highly unlikely... */
@@ -887,13 +942,15 @@ DoItem(int iType)
 		}
 		if (FTPGetCWD(&gConn, gRStartDir, sizeof(gRStartDir)) < 0) {
 			Log(1, "Couldn't get start directory on %s, will try again next time.\n", gHost);
+			(void) AdditionalCmd(&gConn, gPostCommand, NULL);
 			(void) FTPCloseHost(&gConn);
 			return (-1);	/* Try again next time. */
 		}
 		if (gConn.hasCLNT != kCommandNotAvailable)
 			(void) FTPCmd(&gConn, "CLNT NcFTPBatch %.5s %s", gVersion + 11, gOS);
+		(void) AdditionalCmd(&gConn, gPreCommand, NULL);
 
-		if (FTPChdir(&gConn, gRDir) < 0) {
+		if (FTPChdir3(&gConn, gRDir, NULL, 0, cdflags) < 0) {
 			Log(1, "Could not remote cd to %s.\n", gRDir);
 
 			/* Leave open, but unspool.
@@ -913,7 +970,7 @@ DoItem(int iType)
 			return (-1);	/* Try again next time, in case conn dropped. */
 		}
 
-		if (FTPChdir(&gConn, gRDir) < 0) {
+		if (FTPChdir3(&gConn, gRDir, NULL, 0, cdflags) < 0) {
 			Log(1, "Could not remote cd to %s.\n", gRDir);
 			return (-1);	/* Try again next time, in case conn dropped. */
 		}
@@ -938,6 +995,7 @@ DoItem(int iType)
 			Log(1, "Remote item %s is already present locally.\n", gRFile);
 			result = 0;
 		} else {
+			(void) AdditionalCmd(&gConn, gPerFileCommand, gRFile);
 			Log(1, "Done with %s.\n", gRFile);
 		}
 	} else /* if (gOperation == 'P') */ {
@@ -959,6 +1017,7 @@ DoItem(int iType)
 			Log(1, "Local item %s is already present on remote host.\n", gLFile);
 			result = 0;
 		} else {
+			(void) AdditionalCmd(&gConn, gPerFileCommand, gRFile);
 			Log(1, "Done with %s.\n", gLFile);
 		}
 	}
@@ -1032,7 +1091,7 @@ EventShell(volatile unsigned int sleepval)
 {
 	int nItems;
 	struct dirent *direntp;
-	struct stat st;
+	struct Stat st;
 	char *cp;
 	int iType;
 	int iyyyymmdd, ihhmmss, nyyyymmdd, nhhmmss;
@@ -1108,7 +1167,7 @@ EventShell(volatile unsigned int sleepval)
 			(void) STRNCPY(gItemPath, gSpoolDir);
 			(void) STRNCAT(gItemPath, LOCAL_PATH_DELIM_STR);
 			(void) STRNCAT(gItemPath, direntp->d_name);
-			if ((stat(gItemPath, &st) < 0) || (S_ISREG(st.st_mode) == 0)) {
+			if ((Stat(gItemPath, &st) < 0) || (S_ISREG(st.st_mode) == 0)) {
 				/* Item may have been
 				 * deleted by another
 				 * process.
@@ -1159,11 +1218,19 @@ EventShell(volatile unsigned int sleepval)
 					/* rename it back, so it will
 					 * get reprocessed.
 					 */
-					(void) rename(gMyItemPath, gItemPath);
+					if (rename(gMyItemPath, gItemPath) != 0) {
+						/* quit now */
+						Log(0, "Could not rename job %s!\n", gMyItemPath);
+						return;
+					}
 					Log(0, "Re-queueing %s.\n", gItemPath);
 				} else {
-					(void) unlink(gMyItemPath);
 					Log(0, "Done with %s.\n", gItemPath);
+					if (unlink(gMyItemPath) != 0) {
+						/* quit now */
+						Log(0, "Could not delete finished job %s!\n", gMyItemPath);
+						return;
+					}
 				}
 				(void) chdir(LOCAL_PATH_DELIM_STR);
 #if defined(WIN32) || defined(_WINDOWS)
@@ -1175,6 +1242,7 @@ EventShell(volatile unsigned int sleepval)
 			if (gQuitRequested != 0) {
 				(void) closedir(DIRp);
 				Log(0, "User requested close.\n");
+				(void) AdditionalCmd(&gConn, gPostCommand, NULL);
 				(void) FTPCloseHost(&gConn);
 				gMayCancelJmp = 0;
 				Log(0, "-----done-----\n");
@@ -1189,6 +1257,7 @@ EventShell(volatile unsigned int sleepval)
 			break;
 		}
 	}
+	(void) AdditionalCmd(&gConn, gPostCommand, NULL);
 	(void) FTPCloseHost(&gConn);
 	gMayCancelJmp = 0;
 	Log(0, "-----done-----\n");
@@ -1205,7 +1274,7 @@ ListQueue(void)
 {
 	int nItems;
 	struct dirent *direntp;
-	struct stat st;
+	struct Stat st;
 	DIR *DIRp;
 	char *cp;
 	int iyyyymmdd, ihhmmss;
@@ -1227,7 +1296,7 @@ ListQueue(void)
 		(void) STRNCPY(gItemPath, gSpoolDir);
 		(void) STRNCAT(gItemPath, LOCAL_PATH_DELIM_STR);
 		(void) STRNCAT(gItemPath, direntp->d_name);
-		if ((stat(gItemPath, &st) < 0) || (S_ISREG(st.st_mode) == 0)) {
+		if ((Stat(gItemPath, &st) < 0) || (S_ISREG(st.st_mode) == 0)) {
 			/* Item may have been
 			 * deleted by another
 			 * process.
@@ -1607,15 +1676,15 @@ PRead(int sfd, char *const buf0, size_t size, int retry)
 
 	errno = 0;
 	nleft = (int) size;
-	while (1) {
-		nread = read(sfd, buf, nleft);
+	for (;;) {
+		nread = (int) read(sfd, buf, nleft);
 		if (nread <= 0) {
 			if (nread == 0) {
 				/* EOF */
-				nread = size - nleft;
+				nread = (int) size - nleft;
 				return (nread);
 			} else if (errno != EINTR) {
-				nread = size - nleft;
+				nread = (int) size - nleft;
 				if (nread == 0)
 					nread = -1;
 				return (nread);
@@ -1630,7 +1699,7 @@ PRead(int sfd, char *const buf0, size_t size, int retry)
 			break;
 		buf += nread;
 	}
-	nread = size - nleft;
+	nread = (int) size - nleft;
 	return (nread);
 }	/* PRead */
 
@@ -1684,14 +1753,13 @@ Daemon(void)
 			continue;
 		if (gConn.ctrlSocketW == i)
 			continue;
-		if (isatty(1) > 0) {
-			/* Close tty files and replace
-			 * with /dev/null.
-			 */
-			(void) close(i);
-			if (devnull >= 0)
-				(void) dup2(devnull, i);
-		}
+
+		/* Close standard descriptors and replace
+		 * with /dev/null.
+		 */
+		(void) close(i);
+		if (devnull >= 0)
+			(void) dup2(devnull, i);
 	}
 
 	if (devnull >= 0)
@@ -1783,7 +1851,7 @@ main(int argc, const char **const argv)
 	int readcore = -1;
 
 	PreInit();
-	while ((c = getopt(argc, (char **) argv, "|:XDdls:w")) > 0) switch(c) {
+	while ((c = getopt(argc, (char **) argv, "|:XDdlSs:w")) > 0) switch(c) {
 		case 'd':
 			runAsDaemon = 1;
 			break;
@@ -1793,6 +1861,9 @@ main(int argc, const char **const argv)
 		case 'l':
 			listonly = 1;
 			break;
+		case 'S':
+			sleep(15);
+			break;
 		case 's':
 			sleepval = (unsigned int) atoi(optarg);
 			break;
